/**************************************************************************/
/*                                                                        */
/*  Menhir                                                                */
/*                                                                        */
/*  Franois Pottier and Yann Rgis-Gianas, INRIA Rocquencourt            */
/*                                                                        */
/*  Copyright 2005 Institut National de Recherche en Informatique et      */
/*  en Automatique. All rights reserved. This file is distributed         */
/*  under the terms of the GNU Library General Public License, with       */
/*  the special exception on linking described in file LICENSE.           */
/*                                                                        */
/**************************************************************************/

(* This is menhir's standard library. It offers a number of
   parameterized nonterminal definitions, such as options and lists,
   that should be useful in a number of circumstances. *)

%%

(* ------------------------------------------------------------------------- *)
(* Options. *)

(* [option(X)] recognizes either nothing or [X]. It produces a value
   of type ['a option] if [X] produces a value of type ['a]. *)

%public option(X):
  /* nothing */
    { None }
| x = X
    { Some x }

(* [ioption(X)] is identical to [option(X)], except its definition is
   inlined. This has the effect of duplicating the production that
   refers to it, possibly eliminating an LR(1) conflict. *)

%public %inline ioption(X):
  /* nothing */
    { None }
| x = X
    { Some x }

(* [boption(X)] recognizes either nothing or [X]. It produces a value
   of type [bool]. *)

%public boption(X):
  /* nothing */
    { false }
| X
    { true }

(* [loption(X)] recognizes either nothing or [X]. It produces a value
   of type ['a list] if [X] produces a value of type ['a list]. *)

%public loption(X):
  /* nothing */
    { [] }
| x = X
    { x }

(* ------------------------------------------------------------------------- *)
(* Sequences. *)

(* [preceded(opening, X)] recognizes the sequence [opening X]. It
   passes on the value produced by [X], so that it produces a value of
   type ['a] if [X] produces a value of type ['a]. *)

%public %inline preceded(opening, X):
  opening; x = X
    { x }

(* [terminated(X, closing)] recognizes the sequence [X closing]. It
   passes on the value produced by [X], so that it produces a value of
   type ['a] if [X] produces a value of type ['a]. *)

%public %inline terminated(X, closing):
  x = X; closing
    { x }

(* [delimited(opening, X, closing)] recognizes the sequence [opening X
   closing]. It passes on the value produced by [X], so that it
   produces a value of type ['a] if [X] produces a value of type
   ['a]. *)

%public %inline delimited(opening, X, closing):
  opening; x = X; closing
    { x }

(* ------------------------------------------------------------------------- *)
(* Reversed, left-recursive lists. *)

(* [reversed_list(X)] recognizes a possibly empty list of [X]'s. It
   produces a value of type ['a list] if [X] produces a value of type
   ['a]. The front element of the list is the last element that was
   parsed. *)

%public reversed_list(X):
  /* nothing */
    { [] }
| xs = reversed_list(X); x = X
    { x :: xs }

(* [reversed_nonempty_list(X)] recognizes a nonempty list of [X]'s. It
   produces a value of type ['a list] if [X] produces a value of type
   ['a]. The front element of the list is the last element that was
   parsed. *)

%public reversed_nonempty_list(X):
  x = X
    { [ x ] }
| xs = reversed_nonempty_list(X); x = X
    { x :: xs }

(* [reversed_separated_list(separator, X)] recognizes a possibly empty
   list of [X]'s, separated with [separator]'s. It produces a value of
   type ['a list] if [X] produces a value of type ['a]. The front
   element of the list is the last element that was parsed. *)

%public %inline reversed_separated_list(separator, X):
  xs = loption(reversed_separated_nonempty_list(separator, X))
    { xs }

(* [reversed_separated_nonempty_list(separator, X)] recognizes a
   nonempty list of [X]'s, separated with [separator]'s. It produces a
   value of type ['a list] if [X] produces a value of type ['a]. The
   front element of the list is the last element that was parsed. *)

%public reversed_separated_nonempty_list(separator, X):
  x = X
    { [ x ] }
| xs = reversed_separated_nonempty_list(separator, X); separator; x = X
    { x :: xs }

(* ------------------------------------------------------------------------- *)
(* Non-reversed, left-recursive lists. *)

(* [list(X)] recognizes a possibly empty list of [X]'s. It produces a
   value of type ['a list] if [X] produces a value of type ['a]. The
   front element of the list is the first element that was parsed. *)

%public %inline list(X):
  xs = reversed_list(X)
    { List.rev xs }

(* [nonempty_list(X)] recognizes a nonempty list of [X]'s. It produces
   a value of type ['a list] if [X] produces a value of type ['a]. The
   front element of the list is the first element that was parsed. *)

%public %inline nonempty_list(X):
  xs = reversed_nonempty_list(X)
    { List.rev xs }

(* [separated_nonempty_list(separator, X)] recognizes a nonempty list
   of [X]'s, separated with [separator]'s. It produces a value of type
   ['a list] if [X] produces a value of type ['a]. The front element
   of the list is the first element that was parsed. *)

%public %inline separated_nonempty_list(separator, X):
  xs = reversed_separated_nonempty_list(separator, X)
    { List.rev xs }

(* [separated_list(separator, X)] recognizes a possibly empty list of
   [X]'s, separated with [separator]'s. It produces a value of type
   ['a list] if [X] produces a value of type ['a]. The front element
   of the list is the first element that was parsed. *)

%public %inline separated_list(separator, X):
  xs = reversed_separated_list(separator, X)
    { List.rev xs }

%%

